"""
Provider服务到注册中心的注册器
"""
import json
import socket
from sys import exit as sys_exit


class LinkRegister:
    def __init__(self, server_conf):
        """
        初始化
        :param server_conf: 注册中心配置
        """
        self.server_conf = server_conf

    def register_int(self, send_data):
        """
        服务提供者注册
        :return: None
        """
        register = None
        try:
            register = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        except socket.error as e:
            print("[eqlink] [Provider Socket创建] [ERROR]: " + str(e))
            sys_exit()
        try:
            register.connect((self.server_conf['IP'], self.server_conf['PORT']))
            ''' register.setblocking(False) # 设置非阻塞模式 '''
            print('[eqlink] [连接注册中心] [SUCCESS] IP:' + self.server_conf['IP'])
        except socket.gaierror as e:
            print("[eqlink] [连接注册中心] [ERROR]: " + str(e))
            sys_exit()
        '''发送信息到注册中心'''
        try:
            register.sendall(bytes(json.dumps(send_data), encoding="utf8"))
            print("[eqlink] [Provider注册] [register data]:", json.dumps(send_data))
            data = register.recv(self.server_conf['BUF_SIZE'])
            print('[eqlink] [Provider注册] [response]:', str(data, 'UTF-8'))
        except socket.error as e:
            print("[eqlink] [Provider注册] [error]: " + str(e))
            sys_exit()
        ''' 关闭注册连接 '''
        register.close()



"""
yaml配置文件读取工具
"""
import yaml


class ReadConf:
    def __init__(self, path):
        """
        配置文件读取类初始化
        :param path: 配置文件路径
        """
        self.conf_path = path
        self.configuration = {}

    def __read__(self, item):
        """
        读取配置项
        :param item: 要读取的配置项 如：a.b.c
        :return:
        """
        if bool(self.configuration) is False:
            ''' 相同的yaml文件读取一次即可 '''
            try:
                f = open(self.conf_path, 'r', encoding='utf-8')
                cont = f.read()
                self.configuration = yaml.load(cont, Loader=yaml.FullLoader)
            except Exception as e:
                print(e)
                return ''
        ''' 解析要读取的配置项结果 '''
        items = item.split('.')
        res = self.configuration
        for i in items:
            res = res[i]
        return res

    def __file_load__(self):
        """
        加载配置文件内容到 self.configuration
        :return: void
        """
        if bool(self.configuration) is False:
            try:
                f = open(self.conf_path, 'r', encoding='utf-8')
                cont = f.read()
                self.configuration = yaml.load(cont, Loader=yaml.FullLoader)
            except Exception as e:
                print(e)



"""
公共工具，协议分析
"""
from eqlink.components.link_list import LinkList

''' 全局共享：注册中心连接列表 '''
link_list = LinkList()


def protocol_analysis(protocol):
    """
    对协议类型进行分析和处理
    :param protocol: JSON协议数据
    :return: void
    """
    if protocol['type'] == 'provider register':
        ''' Provider注册 '''
        link_list.add_provider(protocol)
        return {'code': '1000', 'message': '服务注册执行完成!'}
    elif protocol['type'] == 'get provider':
        ''' Consumer查询Provider服务列表 '''
        return link_list.provider_list



"""
配置文件的读取
"""
from eqlink.components.read_conf import ReadConf


class LoadConf:
    def __init__(self, path, start_params):
        """
        初始化
        :param path: 配置文件路径
        :param start_params: 应用启动参数
        """
        self.configuration_ext = {}
        self.configuration = ReadConf(path)
        self.configuration.__file_load__()
        if 'env' in start_params:
            ''' 启动参数传入运行环境信息时，额外加载环境配置文件 '''
            self.load_ext(path[:-5] + '-' + start_params['env'] + path[-5:])
        else:
            ''' 启动参数未传入运行环境信息，判断主配置文件是否指定环境信息 '''
            if 'env' in self.configuration.configuration['app']:
                self.load_ext(path[:-5] + '-' + self.configuration.configuration['app']['env'] + path[-5:])
        if 'port' in start_params:
            ''' 支持启动参数强制指定端口 '''
            self.configuration.configuration['app']['link_server']['port'] = int(start_params['port'])

    def node_read(self, conf_node):
        """
        读取服务启动配置信息
        :return: 配置信息节点
        """
        return self.configuration.__read__(conf_node)

    def load_ext(self, path):
        """
        加载附加环境配置文件信息
        :param path: 附加环境路径
        :return: void
        """
        self.configuration_ext = ReadConf(path)
        self.configuration_ext.__file_load__()
        self.configuration.configuration.update(self.configuration_ext.configuration)



"""
远程服务：用于Consumer客户端本地读取Provider列表
"""


class RemoteServer:
    def __init__(self):
        self.remote_server = {}

    def __get__(self):
        return self.remote_server

    def __set__(self, data):
        self.remote_server = data


remote_server = RemoteServer()



"""
子模块读取工具
"""
from importlib import import_module
from sys import path as sys_path


class ReadModule:
    def __init__(self, path=None):
        """
        初始化
        :param path: 要读取的子模块所在路径
        """
        if path is None:
            path = '..'
        sys_path.append(path)
        self.module_list = {}

    def read(self, item):
        """
        要读取的子模块
        :param item: 子模块名称
        :return: 子模块
        """
        if item in self.module_list.keys():
            '''相同的子模块如果已读取，直接返回'''
            return self.module_list[item]
        '''子模块导入'''
        module_item = import_module(item)
        self.module_list[item] = module_item
        return module_item



"""
用于客户端Consumer到注册中心的连接器
"""
import json
import socket
from sys import exit as sys_exit
from time import sleep as time_sleep
from eqlink.components.remote_server import remote_server


class LinkClient:
    def __init__(self, server_conf, client_conf):
        """
        初始化
        :param server_conf: 注册中心配置
        :param client_conf: Consumer客户端配置
        """
        self.server_conf = server_conf
        self.client_conf = client_conf

    def client_int(self, data_to_server):
        """
        客户端 socket 初始化
        :param data_to_server: 发送到注册中心的数据
        :return: void
        """
        client = None
        try:
            ''' 创建socket套接字 '''
            client = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        except socket.error as e:
            print("[eqlink] Error creating socket: %s" + str(e))
            sys_exit()

        try:
            '''  服务端的IP地址和端口 '''
            client.connect((self.server_conf['IP'], self.server_conf['PORT']))
            '''
            setblocking(flag):
                如果flag为0，则将套接字设为非阻塞模式，否则将套接字设为阻塞模式（默认值）。
                非阻塞模式下，如果调用recv()没有发现任何数据，或send()调用无法立即发送数据，那么将引起socket.error异常。
            --- code ---
            client.setblocking(False)
            '''
            print('[eqlink] [连接服务端] [success] IP:' + self.server_conf['IP'])
        except socket.error as e:
            print("[eqlink] connected to server error: %s" + str(e))
            sys_exit()

        ''' 与注册中心保持连接，定时获取服务列表 '''
        while True:
            data_json = json.dumps(data_to_server)
            try:
                client.sendall(bytes(data_json, encoding="utf8"))
                data = client.recv(self.server_conf['BUF_SIZE'])
                ''' 注册中心服务列表写入共享存储区 '''
                remote_server.__set__(json.loads(data))
                print('[eqlink] [获取Provider服务列表]:', str(data, 'UTF-8'))
            except socket.error:
                print("[eqlink] Send failed!!")
                sys_exit()
            ''' 间隔一段时间，进行一次心跳检擦，心跳数据设置 '''
            time_sleep(self.client_conf['alive'])



"""
注册中心本地服务实现
"""

import json
import socket
import sys
from threading import Thread
from eqlink.components.protocol import *


class LinkServer:
    def __init__(self, server_conf):
        self.server_conf = server_conf

    def __data_recv__(self, c_connect):
        """
        等待客户端请求
        :param c_connect: socket 连接
        :return: void
        """
        while True:
            '''处理客户端端数据'''
            try:
                """
                recv(buffer_size) 接收TCP数据，数据以字符串形式返回，buffer_size 指定要接收的最大数据量
                """
                data = c_connect.recv(self.server_conf['BUF_SIZE'])
                data = str(data, 'UTF-8')
                ''' 客户端发送空内容，关闭连接 '''
                # TODO 累计客户都端发送空字符串次数，达到一定次数时才关闭连接
                if data == '':
                    break
                ''' 解析客户端请求，不同类型的请求，进行不同的处理 '''
                data_json = json.loads(data)
                response = protocol_analysis(data_json)
                ''' 响应客户端请求 '''
                c_connect.sendall(bytes(json.dumps(response).encode('utf-8')))
            except socket.error as e:
                print('[eqlink] [Socket Error]', str(e))
                break
        '''关闭客户端连接'''
        c_connect.close()

    def server_init(self):
        """
        服务端 socket 初始化
        :return: None
        """

        try:
            ''' 创建Socket套接字 '''
            server = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        except Exception as e:
            print('[eqlink] [注册中心 socket] [初始化异常] ' + str(e))
            sys.exit()

        try:
            ''' 绑定本地端口 主机地址 '''
            self.server_conf['HOST'] = socket.gethostbyname(socket.gethostname())
            server.bind((self.server_conf['HOST'], self.server_conf['PORT']))
            print('[eqlink] 注册中心绑定 ' + self.server_conf['HOST'] + ':' + str(self.server_conf['PORT']))
        except socket.error as e:
            print("[eqlink] Bind failed!" + str(e))
            sys.exit()

        """
        服务启动 listen(backlog) 开始TCP监听
        backlog指定在拒绝连接之前，操作系统可以挂起的最大连接数量。该值至少为1，大部分应用程序设为5即可
        """
        server.listen(self.server_conf['BACKLOG'])
        print("[eqlink] 初始化已完成 Waiting for connection")

        while True:
            ''' accept() 被动接受TCP客户端连接，(阻塞式)等待连接的到来 '''
            connect, addr = server.accept()
            print("[eqlink] Connected with %s:%s " % (addr[0], str(addr[1])))
            ''' 有客户都端连接后，另启动一个线程 '''
            Thread(target=self.__data_recv__, args=(connect,)).start()



"""
用于存储服务消费者和提供者列表
"""


class LinkList:
    def __init__(self):
        """
        provider_list: 服务提供者列表
        consumer_list: TODO 服务消费者连接列表
        """
        self.provider_list = {}
        self.consumer_list = {}

    def add_provider(self, provider):
        """
        将服务提供者加入到注册中心本地
        :param provider: 服务提供者信息
        :return: void
        """
        service_name = provider['service_name']
        if service_name in self.provider_list:
            '''服务已注册，对比是否增加服务可调用IP或方法'''
            if 'remote' in self.provider_list[service_name]:
                '''如果服务IP不在列表，则添加'''
                if provider['remote'] not in self.provider_list[service_name]['remote']:
                    self.provider_list[service_name]['remote'].append(provider['remote'])
                '''调整服务的方法列表'''
                self.provider_list[service_name]['func'] = provider['func']
        else:
            '''首次进行服务注册'''
            self.provider_list[service_name] = {
                'remote': [provider['remote']],
                'func': provider['func']
            }



from setuptools import setup, find_packages

setup(
    name="eqlink",
    version="0.0.3",
    keywords=("eqlink", "Registration Center", "eqsmart"),
    description="注册中心",
    long_description="增加配置文件加载组件",
    license="GPL-2.0",

    url="https://gitee.com/jingenqiang/eqlink.git",
    author="eq",
    author_email="eq_enqiang@163.com",

    packages=find_packages(),
    include_package_data=True,
    platforms="any",
    install_requires=['PyYAML==6.0']
)

"""
项目打包
python setup.py bdist_egg     # 生成类似 eqlink-0.0.1-py2.7.egg，支持 easy_install 
# 使用此方式
python setup.py sdist         # 生成类似 eqlink-0.0.1.tar.gz，支持 pip
# twine 需要安装
twine upload dist/eqlink-0.0.3.tar.gz
"""
